var http         = require('http');
var util         = require('util');
var stdOutLogger = require('node-logging');
var toolshed     = require('toolshed');

var headerPattern = /^(HTTP\/|Date:|Connection:|Transfer-Encoding:|X-Piwie:|Content-Length:|Content-Type:)/;

// The headers in the response array are unfortunately not nicely
// packaged in a hash, so we have to manually deal with them and parse
// the actual header string
var extractHeaders = function(res) {
  var headers, hash = {};

  // Parse header string
  headers = res._header
    .split('\r\n')
    .map(function(header) {
      if (headerPattern.test(header)) return null;
      if (header === '') return null;
      return header;
    })
    .filter(function(e) { return e !== null; });

  // Convert to hash
  for (var n = 0, l = headers.length; n < l; n++) {
    var header = headers[n].split(': ');
    hash[header[0]] = header[1];
  }

  return hash;
};

/**
 * Initialize the json-loggly module.
 *
 * @method init
 * @params {String} key Loggly key
 * @params {Object} options Configuration options: `pathFilter` - an array of path endpoints that should be ignored when logging. `fieldFilter` - an array of JSON field names that should be filtered out of the logs (optional)
 */
exports.init = function(key, options) {
  var logglyOptions = {
    method : 'POST',
    host   : 'ec2.logs.loggly.com',
    path   : '/inputs/' + key
  };

  options = options || {};
  options.pathFilter = options.pathFilter || [];
  options.fieldFilter = options.fieldFilter || ['password'];
  var fieldFilterLength = options.fieldFilter.length;

  var dbg = function(obj) {
    _log('debug', obj);
  };

  var inf = function(obj) {
    _log('info', obj);
  };

  var war = function(obj) {
    _log('warning', obj);
  };

  var err = function(obj) {
    _log('error', obj);
  };

  // Helper for easy logging of the request
  var request = function(req, res) {
    var level = res.statusCode < 500 ? 'info' : 'error';

    var json = {
      method : req.method,
      path   : req.url,
      status : res.statusCode
    };

    if ('user-agent'    in req.headers) json.userAgent = req.headers['user-agent'];
    if ('_auth'         in req.loggly)  json.auth      = req.loggly._auth;
    if ('_originalBody' in req.loggly)  json.body      = req.loggly._originalBody;

    var headers = extractHeaders(res);

    if ('X-Response-Time' in headers) {
      var ms = parseInt(headers['X-Response-Time']); // remove the 'ms' part for easy searching
      delete headers['X-Response-Time'];             // no need to log this as a header...
      json.responseTime = ms;                        // we'll add a 1st class attribute instead
    }

    if (!toolshed.isEmpty(headers)) json.resHeaders = headers;
    if (res.body)                   json.resBody = res.body;

    var logCache = res.loggly._log;
    if (logCache.dbg.length > 0) json.dbg = logCache.dbg;
    if (logCache.inf.length > 0) json.inf = logCache.inf;
    if (logCache.war.length > 0) json.war = logCache.war;
    if (logCache.err.length > 0) json.err = logCache.err;

    _log(level, json);
  };

  var middleware = function() {
    return function(req, res, next) {
      if ('loggly' in res) console.log('WARNING: res.loggly already exists!');

      req.loggly = {
        logBody: function(obj) {
          this._originalBody = obj;
        },
        logAuth: function(msg) {
          this._auth = msg;
        }
      };

      res.loggly = {
        logBody: function(obj) {
          res.body = obj;
        },
        _log: {
          dbg: [],
          inf: [],
          war: [],
          err: []
        }
      };

      if (options.pathFilter.indexOf(req.url) === -1) {
        var end = res.end;
        res.end = function(chunk, encoding) {
          res.end = end;
          res.end(chunk, encoding);
          request(req, res);
        };
      }

      next();
    };
  };

  var _log = function(level, json) {
    if (typeof(json) === 'string') json = { msg: json };

    json = _filter(json);

    switch (level) {
    case 'debug':
      stdOutLogger.dbg(util.inspect(json, false, null));
      break;
    case 'info':
      stdOutLogger.inf(util.inspect(json, false, null));
      break;
    case 'warning':
      stdOutLogger.err(util.inspect(json, false, null));
      break;
    case 'error':
      stdOutLogger.bad(util.inspect(json, false, null));
    }

    // The loggly service doesn't have different log levels build in,
    // so we just include the level as part of the reqeust
    json.level = level;

    var data = JSON.stringify(json);

    // Use a function, so we can call it again if the loggly request fails
    var retryTimeout = 100;
    var logToLoggly = function() {
      var loggly = http.request(logglyOptions);

      loggly.on('error', function(e) {
        var retry = retryTimeout < 300000; // 5 minutes

        stdOutLogger.err('Could not send log to loggly (' + (retry ? 'retrying' : 'failing') +'): ' + e.message);

        if (retry) {
          setTimeout(logToLoggly, retryTimeout);
          retryTimeout = Math.floor(retryTimeout * 1.5);
        } else {
          throw new Error("Could contact Loggly's servers after several retries");
        }
      });

      loggly.end(data);
    };

    logToLoggly();
  };

  var _filter = function(json) {
    return toolshed.mapJson(json, function(k,v) {
      var match, n;
      for (n = 0; n < fieldFilterLength; n++) {
        match = k.indexOf(options.fieldFilter[n]) !== -1;
        if (match) break;
      }
      return match ? '[FILTERED]' : v;
    });
  };

  return {
    dbg        : dbg,
    inf        : inf,
    war        : war,
    err        : err,
    request    : request,
    middleware : middleware
  };
};
